import asyncio
import sys
import unittest.mock as mock
from collections.abc import Iterator
from typing import Any

import pytest
from httpx import ConnectError

from yatbaf import LongPolling
from yatbaf.exceptions import RequestTimeoutError

MODULE = "yatbaf.long_polling"
module_obj = sys.modules[MODULE]


class AwaitableMock(mock.AsyncMock):

    def __await__(self) -> Iterator[Any]:
        self.await_count += 1
        return iter([])


@pytest.fixture
def asyncio_mock():
    asyncio_module = mock.Mock()
    asyncio_module.gather = mock.AsyncMock()
    asyncio_module.sleep = mock.AsyncMock()
    asyncio_module.wait_for = mock.AsyncMock()
    asyncio_module.CancelledError = asyncio.CancelledError
    return asyncio_module


@pytest.fixture(autouse=True)
def __mock(monkeypatch, asyncio_mock):
    monkeypatch.setattr(module_obj, "asyncio", asyncio_mock)


@pytest.fixture
def bot_mock():
    bot = mock.AsyncMock()
    bot._api_client = mock.Mock()
    bot._api_client.invoke = mock.AsyncMock()
    bot._api_client.close = mock.AsyncMock()
    return bot


@pytest.mark.asyncio
async def test_startup_running():
    polling = LongPolling(None)
    polling._running = True

    await polling._startup()


@pytest.mark.asyncio
async def test_startup():
    polling = LongPolling(None)

    await polling._startup()
    polling._event.clear.assert_called_once()
    assert polling._running


@pytest.mark.asyncio
async def test_startup_func(bot_mock):
    polling = LongPolling(
        bot_mock, on_startup=[
            mock.AsyncMock(),
            mock.AsyncMock(),
        ]
    )
    await polling._startup()
    for func in polling._on_startup:
        func.assert_awaited_once_with(bot_mock)


@pytest.mark.asyncio
async def test_startup_func_error(bot_mock):
    polling = LongPolling(
        bot_mock, on_startup=[
            mock.AsyncMock(side_effect=ValueError()),
        ]
    )
    with pytest.raises(ValueError):
        await polling._startup()
    assert not polling._running


@pytest.mark.asyncio
async def test_shutdown_not_running(bot_mock):
    polling = LongPolling(bot_mock)
    polling._running = False
    await polling._shutdown()
    bot_mock.shutdown.assert_not_awaited()


@pytest.mark.asyncio
async def test_shutdown_func(bot_mock):
    polling = LongPolling(
        bot_mock, on_shutdown=[
            mock.AsyncMock(),
            mock.AsyncMock(),
        ]
    )
    polling._running = True

    await polling._shutdown()
    bot_mock.shutdown.assert_awaited_once()
    for func in polling._on_shutdown:
        func.assert_awaited_once_with(bot_mock)


@pytest.mark.asyncio
async def test_shutdown_func_error(bot_mock):
    polling = LongPolling(
        bot_mock, on_shutdown=[
            mock.AsyncMock(side_effect=ValueError()),
        ]
    )
    polling._running = True
    with pytest.raises(ValueError):
        await polling._shutdown()
    bot_mock.shutdown.assert_awaited_once()


@pytest.mark.asyncio
async def test_get_updates_params(bot_mock):
    polling = LongPolling(bot_mock)
    polling._event.is_set.side_effect = [False, True]
    result = mock.Mock(result=[])
    bot_mock._api_client.invoke.return_value = result
    queue = mock.Mock()

    await polling._get_updates(queue)
    queue.put_nowait.assert_not_called()
    bot_mock._api_client.invoke.assert_awaited_once_with(
        polling._method,
        timeout=(polling._method.timeout + 5),
    )


@pytest.mark.asyncio
async def test_get_updates(bot_mock):
    polling = LongPolling(bot_mock)
    polling._event.is_set.side_effect = [False, False, True]
    result = mock.Mock(result=[])
    bot_mock._api_client.invoke.return_value = result
    queue = mock.Mock()

    await polling._get_updates(queue)
    queue.put_nowait.assert_not_called()
    bot_mock._api_client.invoke.assert_awaited()
    assert bot_mock._api_client.invoke.call_count == 2


@pytest.mark.asyncio
async def test_get_updates_queue(bot_mock, update):
    polling = LongPolling(bot_mock)
    polling._event.is_set.side_effect = [False, True]
    result = mock.Mock(result=[object(), update])
    bot_mock._api_client.invoke.return_value = result
    queue = mock.Mock()

    await polling._get_updates(queue)
    queue.put_nowait.assert_called_once_with(result.result)
    bot_mock._api_client.invoke.assert_awaited_once()
    assert polling._method.offset == update.update_id + 1
    assert bot_mock._api_client.invoke.call_count == 1


@pytest.mark.asyncio
async def test_get_updates_timeout_error(bot_mock, asyncio_mock):
    bot_mock._error_handler = error_handler = mock.AsyncMock()
    polling = LongPolling(bot_mock)
    polling._event.is_set.side_effect = [False, False, True]
    result = mock.Mock(result=[])
    bot_mock._api_client.invoke.side_effect = [
        result, RequestTimeoutError("", None), result
    ]
    queue = mock.Mock()

    await polling._get_updates(queue)
    error_handler.on_error.assert_not_awaited()
    asyncio_mock.sleep.assert_awaited_once_with(polling._timeout_delay)
    assert bot_mock._api_client.invoke.call_count == 2


@pytest.mark.asyncio
async def test_get_updates_connection_error(bot_mock, asyncio_mock):
    bot_mock._error_handler = error_handler = mock.AsyncMock()
    polling = LongPolling(bot_mock)
    polling._event.is_set.side_effect = [False, False, True]
    result = mock.Mock(result=[])
    bot_mock._api_client.invoke.side_effect = [result, ConnectError(""), result]
    queue = mock.Mock()

    await polling._get_updates(queue)
    error_handler.on_error.assert_not_awaited()
    asyncio_mock.sleep.assert_awaited_once_with(polling._connection_delay)
    assert bot_mock._api_client.invoke.call_count == 2


@pytest.mark.asyncio
@pytest.mark.parametrize("result", [False, True])
async def test_get_updates_error(bot_mock, result):
    polling = LongPolling(
        bot_mock,
        on_error=(error_handler := mock.AsyncMock(return_value=result)),
    )
    polling._event = mock.Mock()
    polling._event.is_set.side_effect = [False, True]
    bot_mock._api_client.invoke.side_effect = exc = ValueError()
    queue = mock.Mock()
    await polling._get_updates(queue)
    error_handler.assert_awaited_once_with(exc)

    if result:
        polling._event.set.assert_not_called()
    else:
        polling._event.set.assert_called_once()


@pytest.mark.asyncio
async def test_main_loop(monkeypatch, bot_mock, asyncio_mock, update):
    tg = mock.AsyncMock()
    tg.__aenter__.return_value = tg
    tg.create_task = mock.Mock()
    asyncio_mock.TaskGroup.return_value = tg

    asyncio_mock.Queue.return_value = queue = mock.Mock()
    monkeypatch.setattr(  # yapf: disable
        LongPolling,
        "_get_updates",
        get_updates := mock.Mock(
            return_value=(
                get_updates_coro := object()
            )
        )
    )
    polling = LongPolling(bot_mock)
    polling._event.is_set.side_effect = [False, True]
    asyncio_mock.create_task.side_effect = [
        get_updates_task := AwaitableMock(cancel=mock.Mock()),
    ]
    bot_mock.process_update = mock.Mock(return_value=(coro := object()))
    asyncio_mock.wait_for.return_value = [update]
    await polling._main_loop()

    get_updates.assert_called_once_with(queue)
    asyncio_mock.create_task.assert_called_once_with(
        get_updates_coro,
        name="_get_updates",
    )
    get_updates_task.cancel.assert_called_once()
    get_updates_task.assert_awaited_once()
    tg.create_task.assert_called_once_with(coro, name=mock.ANY)


@pytest.mark.asyncio
async def test_run(monkeypatch):
    monkeypatch.setattr(LongPolling, "_startup", startup := mock.AsyncMock())
    monkeypatch.setattr(LongPolling, "_main_loop", loop := mock.AsyncMock())
    monkeypatch.setattr(LongPolling, "_shutdown", shutdown := mock.AsyncMock())

    polling = LongPolling(None)
    await polling._run()
    startup.assert_awaited_once()
    loop.assert_awaited_once()
    shutdown.assert_awaited_once()


@pytest.mark.asyncio
async def test_run_startup_error(monkeypatch):
    monkeypatch.setattr(
        LongPolling,
        "_startup",
        startup := mock.AsyncMock(side_effect=ValueError())
    )
    monkeypatch.setattr(LongPolling, "_main_loop", loop := mock.AsyncMock())
    monkeypatch.setattr(LongPolling, "_shutdown", shutdown := mock.AsyncMock())

    polling = LongPolling(None)
    await polling._run()
    assert not polling._running
    startup.assert_awaited_once()
    loop.assert_not_awaited()
    shutdown.assert_not_awaited()


def test_start(monkeypatch, asyncio_mock):
    monkeypatch.setattr(
        LongPolling,
        "_run",
        mock.Mock(return_value=(coro := mock.Mock())),
    )
    LongPolling(None).start()
    asyncio_mock.run.assert_called_once_with(coro)


def test_stop():
    polling = LongPolling(None)
    polling.stop()
    polling._event.set.assert_called_once()



from __future__ import annotations

from typing import TYPE_CHECKING
from typing import ClassVar
from typing import final

from yatbaf.types import Message

from .abc import TelegramMethodWithMedia

if TYPE_CHECKING:
    from yatbaf.types import InputMediaAudio
    from yatbaf.types import InputMediaDocument
    from yatbaf.types import InputMediaPhoto
    from yatbaf.types import InputMediaVideo
    from yatbaf.types import ReplyParameters
    from yatbaf.typing import NoneBool
    from yatbaf.typing import NoneInt
    from yatbaf.typing import NoneStr


@final
class SendMediaGroup(TelegramMethodWithMedia[list[Message]]):
    """See :meth:`yatbaf.bot.Bot.send_media_group`"""

    chat_id: str | int
    # yapf: disable
    media: list[InputMediaAudio | InputMediaDocument | InputMediaPhoto | InputMediaVideo]  # noqa: E501
    # yapf: enable
    business_connection_id: NoneStr = None
    message_thread_id: NoneInt = None
    disable_notification: NoneBool = None
    protect_content: NoneBool = None
    reply_parameters: ReplyParameters | None = None

    __meth_media_fields__: ClassVar[tuple[str, ...]] = ("media",)



__all__ = (
    "BaseFilter",
    "Chat",
    "ChatId",
    "Command",
    "Content",
    "Text",
    "User",
    "text",
    "video",
    "document",
    "media",
    "audio",
    "photo",
    "sticker",
    "private",
    "group",
    "channel"
)

from .base import BaseFilter
from .chat_id import Chat
from .chat_id import ChatId
from .chat_id import channel
from .chat_id import group
from .chat_id import private
from .command import Command
from .content_type import Content
from .content_type import audio
from .content_type import document
from .content_type import media
from .content_type import photo
from .content_type import sticker
from .content_type import text
from .content_type import video
from .text_content import Text
from .user import User

Command.incompat(Command)
Command.incompat(Text)
Command.incompat(Content)

Text.incompat(Content)

User.incompat(User)

Chat.incompat(Chat)

ChatId.incompat(ChatId)

Content.incompat(Content)



def test_ctx_empty(message):
    assert message.__usrctx__ == {"ctx": {}}
    assert not message.ctx


def test_ctx(message):
    message.ctx["foo"] = "bar"
    assert "foo" in message.__usrctx__["ctx"]
    assert message.__usrctx__["ctx"]["foo"] == "bar"



import pytest

from yatbaf.filters import Command
from yatbaf.filters import text
from yatbaf.handler import Handler
from yatbaf.handler import on_message
from yatbaf.router import OnMessage


@pytest.fixture(autouse=True)
def __reset_update(message):
    message.ctx["test"] = []


def create_handler(mark, filters=None):

    @on_message(filters=filters)
    async def handler(update):
        update.ctx["test"].append(mark)

    return handler


@pytest.mark.asyncio
async def test_fallback(message):
    router = OnMessage(handlers=[create_handler("f")])
    router.on_registration()

    await router.handle(message)
    assert message.ctx["test"] == ["f"]


@pytest.mark.asyncio
async def test_resolve_filters(message):
    router = OnMessage(
        handlers=[
            create_handler("f"),
            create_handler("foo", filters=[Command("foo")]),
        ]
    )
    router.on_registration()
    message.text = "/foo"

    await router.handle(message)
    assert message.ctx["test"] == ["foo"]


@pytest.mark.asyncio
async def test_resolve_stop_propagate(message):
    router = OnMessage(
        handlers=[
            create_handler("foo", filters=[Command("foo")]),
            create_handler("foo1", filters=[Command("foo")]),
        ],
        routers=[
            OnMessage(
                handlers=[create_handler("foo2", filters=[Command("foo")])],
            ),
        ],
    )
    router.on_registration()
    message.text = "/foo"

    await router.handle(message)
    assert message.ctx["test"] == ["foo"]


@pytest.mark.asyncio
async def test_resolve_nested_vert(message):
    router = OnMessage(
        handlers=[
            create_handler("bar1", filters=[Command("bar")]),
        ],
        routers=[
            OnMessage(
                handlers=[
                    create_handler("bar2", filters=[Command("bar")]),
                ],
                routers=[
                    OnMessage(
                        handlers=[
                            create_handler("foo", filters=[Command("foo")]),
                        ],
                        routers=[
                            OnMessage(
                                handlers=[
                                    create_handler(
                                        "bar3", filters=[Command("bar")]
                                    ),
                                ]
                            ),
                        ],
                    ),
                ],
            ),
        ],
    )
    router.on_registration()
    message.text = "/foo"

    await router.handle(message)
    assert message.ctx["test"] == ["foo"]


@pytest.mark.asyncio
async def test_resolve_nested_horiz(message):
    router = OnMessage(
        routers=[
            OnMessage(
                handlers=[
                    create_handler("foo", filters=[Command("foo")]),
                ]
            ),
            OnMessage(
                handlers=[
                    create_handler("bar", filters=[Command("bar")]),
                ]
            ),
            OnMessage(
                handlers=[
                    create_handler("foo1", filters=[Command("foo")]),
                ]
            ),
        ],
    )
    router.on_registration()
    message.text = "/foo"

    await router.handle(message)
    assert message.ctx["test"] == ["foo"]


@pytest.mark.asyncio
async def test_stop_propagate_true_param(message, handler_fn, mock_mark):
    router = OnMessage(
        routers=[
            OnMessage(stop_propagate=True),
            OnMessage(handlers=[Handler(handler_fn, "message")]),
        ],
    )
    router.on_registration()
    await router.handle(message)
    mock_mark.assert_not_called()


@pytest.mark.asyncio
async def test_stop_propagate_true_filter(message, handler_fn, mock_mark):
    router = OnMessage(
        routers=[
            OnMessage(filters=[text]),
            OnMessage(handlers=[Handler(handler_fn, "message")]),
        ],
    )
    router.on_registration()
    message.text = "foo"
    await router.handle(message)
    mock_mark.assert_not_called()


@pytest.mark.asyncio
async def test_stop_propagate_true_filter_guard(
    message, handler_fn, mock_mark, guard_false
):
    router = OnMessage(
        routers=[
            OnMessage(filters=[text], guards=[guard_false]),
            OnMessage(handlers=[Handler(handler_fn, "message")]),
        ],
    )
    router.on_registration()
    message.text = "foo"
    await router.handle(message)
    mock_mark.assert_not_called()


@pytest.mark.asyncio
async def test_stop_propagate_false_filter(message, handler_fn, mock_mark):
    router = OnMessage(
        routers=[
            OnMessage(filters=[text], stop_propagate=False),
            OnMessage(handlers=[Handler(handler_fn, "message")]),
        ],
    )
    router.on_registration()
    message.text = "foo"
    await router.handle(message)
    mock_mark.assert_called_once()



from __future__ import annotations

from typing import TYPE_CHECKING
from typing import final

from yatbaf.types import Message

from .abc import TelegramMethod

if TYPE_CHECKING:
    from yatbaf.types import ReplyParameters
    from yatbaf.typing import NoneBool
    from yatbaf.typing import NoneInt
    from yatbaf.typing import NoneStr
    from yatbaf.typing import ReplyMarkup


@final
class SendContact(TelegramMethod[Message]):
    """See :meth:`yatbaf.bot.Bot.send_contact`"""

    chat_id: str | int
    phone_number: str
    first_name: str
    business_connection_id: NoneStr = None
    message_thread_id: NoneInt = None
    last_name: NoneStr = None
    vcard: NoneStr = None
    disable_notification: NoneBool = None
    protect_content: NoneBool = None
    reply_parameters: ReplyParameters | None = None
    reply_markup: ReplyMarkup | None = None



import pytest

from yatbaf.enums import UpdateType
from yatbaf.filters import Command
from yatbaf.handler import Handler
from yatbaf.router import OnMessage


@pytest.mark.asyncio
async def test_empty_router(message):
    router = OnMessage()
    router.on_registration()
    assert await router._find_handler(message) is None


@pytest.mark.asyncio
async def test_handler_no_filters(message, handler_fn):
    handler = Handler(handler_fn, update_type="message")
    router = OnMessage(handlers=[handler])
    router.on_registration()
    assert await router._find_handler(message) is handler


@pytest.mark.asyncio
async def test_fallback_handler(message, handler_fn):
    fallback = Handler(handler_fn, update_type=UpdateType.MESSAGE)
    router = OnMessage(handlers=[fallback])

    @router(filters=[Command("foo")])
    async def cmd_foo(_):
        pass

    router.on_registration()
    message.text = "/bar"
    assert await router._find_handler(message) is fallback



from __future__ import annotations

from typing import TYPE_CHECKING
from typing import final

from yatbaf.types import Message

from .abc import TelegramMethod

if TYPE_CHECKING:
    from yatbaf.types import InlineKeyboardMarkup
    from yatbaf.typing import NoneInt
    from yatbaf.typing import NoneStr


@final
class StopMessageLiveLocation(TelegramMethod[Message | bool]):
    """See :meth:`yatbaf.bot.Bot.stop_message_live_location`"""

    chat_id: str | int | None = None
    message_id: NoneInt = None
    inline_message_id: NoneStr = None
    reply_markup: InlineKeyboardMarkup | None = None



import pytest

from yatbaf.dispatcher import Dispatcher
from yatbaf.handler import Handler
from yatbaf.router import OnMessage
from yatbaf.router import OnPoll


@pytest.mark.asyncio
async def test_resolve(handler_fn, update, mock_mark):
    router = OnMessage(handlers=[Handler(handler_fn, "message")])
    router.on_registration()

    dispatcher = Dispatcher(handlers={"message": router})
    await dispatcher.resolve(update)
    mock_mark.assert_called_once_with(update.event)


@pytest.mark.asyncio
async def test_resolve_none(handler_fn, update, mock_mark):
    dispatcher = Dispatcher(
        handlers={"poll": OnPoll(handlers=[Handler(handler_fn, "poll")])}
    )
    await dispatcher.resolve(update)
    mock_mark.assert_not_called()



from __future__ import annotations

from typing import TYPE_CHECKING
from typing import final

from .abc import TelegramMethod

if TYPE_CHECKING:
    from yatbaf.types import ChatPermissions
    from yatbaf.typing import NoneBool


@final
class SetChatPermissions(TelegramMethod[bool]):
    """See :meth:`yatbaf.bot.Bot.set_chat_permissions`"""

    chat_id: str | int
    permissions: ChatPermissions
    use_independent_chat_permissions: NoneBool = None



import pytest

from yatbaf.exceptions import BotWarning
from yatbaf.exceptions import FrozenInstanceError
from yatbaf.handler import Handler
from yatbaf.router import OnMessage
from yatbaf.router import OnPoll


def test_init():
    router = OnMessage(routers=[nested := OnMessage()])
    assert router._routers == [nested]
    assert nested.parent is router


def test_add_router():
    router = OnMessage()
    nested = OnMessage()
    router.add_router(nested)
    assert router._routers == [nested]
    assert nested.parent is router


def test_add_router_self():
    router = OnMessage()
    with pytest.raises(ValueError):
        router.add_router(router)


def test_add_router_wrong_type():
    router = OnMessage()
    with pytest.raises(ValueError):
        router.add_router(OnPoll())


def test_add_router_registered():
    _ = OnMessage(
        routers=[
            nested := OnMessage(),
        ],
    )
    router2 = OnMessage()
    with pytest.raises(ValueError):
        router2.add_router(nested)


def test_add_router_duplicate_same_obj():
    router = OnMessage(
        routers=[
            nested := OnMessage(),
        ],
    )
    with pytest.warns(BotWarning):
        router.add_router(nested)
    assert router._routers == [nested]


def test_add_router_duplicate_equal_obj():

    async def fn(_):
        pass

    router = OnMessage(
        routers=[nested := OnMessage(handlers=[Handler(fn, "message")])]
    )
    dup = OnMessage(handlers=[Handler(fn, "message")])
    with pytest.warns(BotWarning):
        router.add_router(dup)
    assert router._routers == [nested]


def test_add_router_frozen():
    router = OnMessage()
    router.on_registration()
    with pytest.raises(FrozenInstanceError):
        router.add_router(OnMessage())



from __future__ import annotations

from typing import final

from .abc import TelegramMethod


@final
class SetChatTitle(TelegramMethod[bool]):
    """See :meth:`yatbaf.bot.Bot.set_chat_title`"""

    chat_id: str | int
    title: str



import pytest

from yatbaf.di import Provide
from yatbaf.di import create_dependency_batches
from yatbaf.di import create_dependency_graph
from yatbaf.di import resolve_dependencies


async def provide_error():
    raise ValueError


async def provide_data():
    return "data"


dependencies = {
    "data": Provide(provide_data),
    "error": Provide(provide_error),
}


@pytest.mark.asyncio
async def test_exception_single_dep():
    error_graph = create_dependency_graph("error", dependencies)
    batches = create_dependency_batches({error_graph})
    with pytest.raises(ValueError):
        await resolve_dependencies(batches, {})


@pytest.mark.asyncio
async def test_exception_multi_deps():
    error_graph = create_dependency_graph("error", dependencies)
    data_graph = create_dependency_graph("data", dependencies)
    batches = create_dependency_batches({data_graph, error_graph})
    with pytest.raises(ValueError):
        await resolve_dependencies(batches, {})



import pytest

from yatbaf.filters.base import BaseFilter


class FalseFilter(BaseFilter):
    priority = {"content": (1, 100)}

    async def check(self, _):
        return False


class TrueFilter(BaseFilter):
    priority = {"content": (1, 100)}

    async def check(self, _):
        return True


@pytest.mark.asyncio
async def test_not_true(message):
    assert await (~FalseFilter()).check(message)


@pytest.mark.asyncio
async def test_not_false(message):
    assert not await (~TrueFilter()).check(message)


@pytest.mark.asyncio
async def test_or_true(message):
    assert await (FalseFilter() | TrueFilter()).check(message)


@pytest.mark.asyncio
async def test_or_false(message):
    assert not await (FalseFilter() | FalseFilter()).check(message)


@pytest.mark.asyncio
async def test_and_true(message):
    assert await (TrueFilter() & TrueFilter()).check(message)


@pytest.mark.asyncio
async def test_and_false(message):
    assert not await (TrueFilter() & FalseFilter()).check(message)



import pytest

from yatbaf import filters as f
from yatbaf.exceptions import FilterCompatError
from yatbaf.filters.base import check_compatibility
from yatbaf.filters.base import is_compat


def test_validate():
    with pytest.raises(FilterCompatError):
        is_compat(f.Chat("group"), f.Command("boo") | f.Chat("group"))


@pytest.mark.parametrize(
    "filters",
    [  # yapf: disable
        [f.Command("foo"), f.Command("bar")],
        [f.Command("foo"), f.Text(startswith="bar")],
        [f.Text(startswith="bar"), f.Command("foo")],
        [f.Content("text"), f.Content("document")],
        [f.User(123), f.User(321)],
        [f.User(123), f.User(321) & f.Command("foo")],
        [f.User(321) & f.Command("foo"), f.User(123)],
        [f.User(123) & f.Chat("group"), f.User(321)],
        [f.User(123) & (f.Chat("private") | f.Chat("group")), f.Command("foo"), ~f.User(321)],  # noqa: E501
        [f.Chat("group"), f.Chat("private")],
        [f.Chat("group") & f.Chat("private")],
        [~f.Chat("group"), f.Chat("private")],
        [f.Content("audio") & f.Chat("private"), f.Command("foo")],
        [f.Content("audio") & f.Chat("private") & f.Command("foo")],
        [f.Chat("private"), f.User(123), f.Command("foo"), f.Content("text")],
        [(f.Chat("private") & f.User(123)) & f.User(321)],
        [f.Chat("group"), f.Text(startswith="foo") & f.Text(endswith="bar"), f.Content("text")],   # noqa: E501
    ]
)
def test_conflicts(filters):
    with pytest.raises(FilterCompatError):
        check_compatibility(filters, False)


@pytest.mark.parametrize(
    "filters",
    [  # yapf: disable
        [f.Command("foo"), f.User(123)],
        [f.Text(startswith="bar"), f.Text(endswith="foo")],
        [f.Command("foo"), (f.Chat("private") & f.ChatId(123)) | f.User(123)],
        [f.User(123), f.Chat("group"), f.Command("foo")],
        [f.Content("text") | f.Content("document"), f.Chat("private")],
        [~(f.Content("text") | f.Content("document")), f.Chat("private")],
    ]
)
def test_no_conflicts(filters):
    check_compatibility(filters, False)



import pytest

from yatbaf.dispatcher import Dispatcher
from yatbaf.handler import Handler


@pytest.mark.asyncio
async def test_resolve_guard_false(guard_false, handler_fn, update, mock_mark):

    dispatcher = Dispatcher(
        handlers={"message": Handler(handler_fn, "message")},
        guards=[guard_false],
    )
    await dispatcher.resolve(update)
    mock_mark.assert_not_called()


@pytest.mark.asyncio
async def test_resolve_guard_true(guard_true, handler_fn, update, mock_mark):
    dispatcher = Dispatcher(
        handlers={"message": Handler(handler_fn, "message")},
        guards=[guard_true],
    )
    await dispatcher.resolve(update)
    mock_mark.assert_called_once_with(update.event)



from __future__ import annotations

from typing import final

from yatbaf.typing import NoneBool
from yatbaf.typing import NoneInt
from yatbaf.typing import NoneStr

from ..types import chat  # noqa: F401
from .abc import TelegramType
from .user import User


@final
class GiveawayWinners(TelegramType):
    """This object represents a message about the completion of a giveaway with
    public winners.

    See: https://core.telegram.org/bots/api#giveawaywinners
    """

    chat: chat.Chat
    """The chat that created the giveaway."""

    giveaway_message_id: int
    """Identifier of the messsage with the giveaway in the chat."""

    winners_selection_date: int
    """Point in time (Unix timestamp) when winners of the giveaway were selected."""  # noqa: E501

    winner_count: int
    """Total number of winners in the giveaway."""

    winners: list[User]
    """List of up to 100 winners of the giveaway."""

    additional_chat_count: NoneInt = None
    """*Optional.* The number of other chats the user had to join in order to be
    eligible for the giveaway.
    """

    premium_subscription_month_count: NoneInt = None
    """*Optional.* The number of months the Telegram Premium subscription won
    from the giveaway will be active for.
    """

    unclaimed_prize_count: NoneInt = None
    """*Optional.* Number of undistributed prizes."""

    only_new_members: NoneBool = None
    """*Optional.* ``True``, if only users who had joined the chats after the
    giveaway started were eligible to win.
    """

    was_refunded: NoneBool = None
    """*Optional.* ``True``, if the giveaway was canceled because the payment
    for it was refunded.
    """

    prize_description: NoneStr = None
    """*Optional.* Description of additional giveaway prize."""



from __future__ import annotations

from typing import final

from .abc import TelegramType
from .location import Location


@final
class BusinessLocation(TelegramType):
    """See: https://core.telegram.org/bots/api#businesslocation"""

    address: str
    """Address of the business."""

    location: Location | None = None
    """*Optional.* Location of the business."""



from __future__ import annotations

from typing import final

from .abc import TelegramType


@final
class BusinessOpeningHoursInterval(TelegramType):
    """See: https://core.telegram.org/bots/api#businessopeninghoursinterval"""

    opening_minute: int
    """The minute's sequence number in a week, starting on Monday, marking the
    start of the time interval during which the business is open; 0 - 7 * 24 * 60
    """  # noqa: E501

    closing_minute: int
    """The minute's sequence number in a week, starting on Monday, marking the
    end of the time interval during which the business is open; 0 - 8 * 24 * 60
    """



import pytest

from yatbaf.di import Provide
from yatbaf.enums import UpdateType
from yatbaf.handler import Handler
from yatbaf.router import OnBusinessConnection
from yatbaf.router import OnBusinessMessage
from yatbaf.router import OnCallbackQuery
from yatbaf.router import OnChannelPost
from yatbaf.router import OnChatBoost
from yatbaf.router import OnChatJoinRequest
from yatbaf.router import OnChatMember
from yatbaf.router import OnChosenInlineResult
from yatbaf.router import OnDeletedBusinessMessages
from yatbaf.router import OnEditedBusinessMessage
from yatbaf.router import OnEditedChannelPost
from yatbaf.router import OnEditedMessage
from yatbaf.router import OnInlineQuery
from yatbaf.router import OnMessage
from yatbaf.router import OnMessageReaction
from yatbaf.router import OnMessageReactionCount
from yatbaf.router import OnMyChatMember
from yatbaf.router import OnPoll
from yatbaf.router import OnPollAnswer
from yatbaf.router import OnPreCheckoutQuery
from yatbaf.router import OnRemovedChatBoost
from yatbaf.router import OnShippingQuery
from yatbaf.router import parse_handlers


async def fn(_):
    pass


@pytest.mark.parametrize(
    "handler,exp",
    [
        [
            Handler(fn, UpdateType.MESSAGE), {
                "message": OnMessage(
                    handlers=[Handler(fn, UpdateType.MESSAGE)]
                )
            }
        ],
        [
            Handler(fn, UpdateType.EDITED_MESSAGE),
            {
                "edited_message": OnEditedMessage(
                    handlers=[Handler(fn, UpdateType.EDITED_MESSAGE)]
                )
            }
        ],
        [
            Handler(fn, UpdateType.CHANNEL_POST),
            {
                "channel_post": OnChannelPost(
                    handlers=[Handler(fn, UpdateType.CHANNEL_POST)]
                )
            }
        ],
        [
            Handler(fn, UpdateType.EDITED_CHANNEL_POST),
            {
                "edited_channel_post": OnEditedChannelPost(
                    handlers=[Handler(fn, UpdateType.EDITED_CHANNEL_POST)]
                )
            }
        ],
        [
            Handler(fn, UpdateType.INLINE_QUERY),
            {
                "inline_query": OnInlineQuery(
                    handlers=[Handler(fn, UpdateType.INLINE_QUERY)]
                )
            }
        ],
        [
            Handler(fn, UpdateType.CHOSEN_INLINE_RESULT),
            {
                "chosen_inline_result": OnChosenInlineResult(
                    handlers=[Handler(fn, UpdateType.CHOSEN_INLINE_RESULT)]
                )
            }
        ],
        [
            Handler(fn, UpdateType.CALLBACK_QUERY),
            {
                "callback_query": OnCallbackQuery(
                    handlers=[Handler(fn, UpdateType.CALLBACK_QUERY)]
                )
            }
        ],
        [
            Handler(fn, UpdateType.SHIPPING_QUERY),
            {
                "shipping_query": OnShippingQuery(
                    handlers=[Handler(fn, UpdateType.SHIPPING_QUERY)]
                )
            }
        ],
        [
            Handler(fn, UpdateType.PRE_CHECKOUT_QUERY),
            {
                "pre_checkout_query": OnPreCheckoutQuery(
                    handlers=[Handler(fn, UpdateType.PRE_CHECKOUT_QUERY)]
                )
            }
        ],
        [
            Handler(fn, UpdateType.POLL), {
                "poll": OnPoll(handlers=[Handler(fn, UpdateType.POLL)])
            }
        ],
        [
            Handler(fn, UpdateType.POLL_ANSWER),
            {
                "poll_answer": OnPollAnswer(
                    handlers=[Handler(fn, UpdateType.POLL_ANSWER)]
                )
            }
        ],
        [
            Handler(fn, UpdateType.MY_CHAT_MEMBER),
            {
                "my_chat_member": OnMyChatMember(
                    handlers=[Handler(fn, UpdateType.MY_CHAT_MEMBER)]
                )
            }
        ],
        [
            Handler(fn, UpdateType.CHAT_MEMBER),
            {
                "chat_member": OnChatMember(
                    handlers=[Handler(fn, UpdateType.CHAT_MEMBER)]
                )
            }
        ],
        [
            Handler(fn, UpdateType.CHAT_JOIN_REQUEST),
            {
                "chat_join_request": OnChatJoinRequest(
                    handlers=[Handler(fn, UpdateType.CHAT_JOIN_REQUEST)]
                )
            }
        ],
        [
            Handler(fn, UpdateType.MESSAGE_REACTION),
            {
                "message_reaction": OnMessageReaction(
                    handlers=[Handler(fn, UpdateType.MESSAGE_REACTION)]
                )
            }
        ],
        [
            Handler(fn, UpdateType.MESSAGE_REACTION_COUNT),
            {
                "message_reaction_count": OnMessageReactionCount(
                    handlers=[Handler(fn, UpdateType.MESSAGE_REACTION_COUNT)]
                )
            }
        ],
        [
            Handler(fn, UpdateType.CHAT_BOOST),
            {
                "chat_boost": OnChatBoost(
                    handlers=[Handler(fn, UpdateType.CHAT_BOOST)]
                )
            }
        ],
        [
            Handler(fn, UpdateType.REMOVED_CHAT_BOOST),
            {
                "removed_chat_boost": OnRemovedChatBoost(
                    handlers=[Handler(fn, UpdateType.REMOVED_CHAT_BOOST)]
                )
            }
        ],
        [
            Handler(fn, UpdateType.BUSINESS_CONNECTION),
            {
                "business_connection": OnBusinessConnection(
                    handlers=[Handler(fn, UpdateType.BUSINESS_CONNECTION)]
                )
            }
        ],
        [
            Handler(fn, UpdateType.BUSINESS_MESSAGE),
            {
                "business_message": OnBusinessMessage(
                    handlers=[Handler(fn, UpdateType.BUSINESS_MESSAGE)]
                )
            }
        ],
        [
            Handler(fn, UpdateType.EDITED_BUSINESS_MESSAGE),
            {
                "edited_business_message": OnEditedBusinessMessage(
                    handlers=[Handler(fn, UpdateType.EDITED_BUSINESS_MESSAGE)]
                )
            }
        ],
        [
            Handler(fn, UpdateType.DELETED_BUSINESS_MESSAGES),
            {
                "deleted_business_messages": OnDeletedBusinessMessages(
                    handlers=[
                        Handler(fn, UpdateType.DELETED_BUSINESS_MESSAGES)
                    ]
                )
            }
        ],
    ]
)
def test_parse(handler, exp):
    assert parse_handlers(handlers=[handler]) == exp


def test_one_router(handler_fn):
    router = OnMessage(handlers=[Handler(handler_fn, "message")])
    result = parse_handlers(routers=[router])
    assert result == {"message": router}
    assert result["message"] is router


def test_routers(handler_fn):
    result = parse_handlers(
        routers=[
            OnMessage(),
            OnMessage(handlers=[Handler(handler_fn, "message")]),
            OnPoll(handlers=[Handler(handler_fn, "poll")]),
            OnCallbackQuery(handlers=[Handler(handler_fn, "callback_query")])
        ]
    )
    assert result == {
        "message": OnMessage(
            routers=[
                OnMessage(),
                OnMessage(handlers=[Handler(handler_fn, "message")],)
            ],
        ),
        "callback_query": OnCallbackQuery(
            handlers=[Handler(handler_fn, "callback_query")]
        ),
        "poll": OnPoll(handlers=[Handler(handler_fn, "poll")]),
    }


def test_handlers_routers(handler_fn):
    result = parse_handlers(
        handlers=[
            Handler(handler_fn, "poll"),
            Handler(handler_fn, "message"),
            Handler(handler_fn, "callback_query"),
        ],
        routers=[
            OnMessage(handlers=[Handler(handler_fn, "message")]),
            OnCallbackQuery(handlers=[Handler(handler_fn, "callback_query")]),
        ],
    )
    assert result == {
        "message": OnMessage(
            handlers=[Handler(handler_fn, "message")],
            routers=[OnMessage(handlers=[Handler(handler_fn, "message")])],
        ),
        "poll": OnPoll(handlers=[Handler(handler_fn, "poll")]),
        "callback_query": OnCallbackQuery(
            handlers=[Handler(handler_fn, "callback_query")],
            routers=[
                OnCallbackQuery(
                    handlers=[Handler(handler_fn, "callback_query")]
                )
            ]
        ),
    }


async def provide_data1():
    return 1


async def provide_data2():
    return 2


@pytest.mark.parametrize(
    "gd,rd,exp",
    [  # yapf: disable
        [{}, {}, {}],
        [
            {"data1": Provide(provide_data1)},
            {},
            {"data1": Provide(provide_data1)},
        ],
        [
            {},
            {"data1": Provide(provide_data1)},
            {"data1": Provide(provide_data1)},
        ],
        [
            {"data1": Provide(provide_data1)},
            {"data1": Provide(provide_data2)},
            {"data1": Provide(provide_data2)},
        ],
        [
            {"data1": Provide(provide_data1)},
            {"data2": Provide(provide_data2)},
            {
                "data1": Provide(provide_data1),
                "data2": Provide(provide_data2),
            },
        ],
    ]
)
def test_deps_one_router_merge(gd, rd, exp):
    router = OnMessage(dependencies=rd)
    result = parse_handlers(routers=[router], dependencies=gd)
    assert result == {"message": OnMessage(dependencies=exp)}


def test_deps_handler(handler_fn):
    dependencies = {"data": Provide(provide_data1)}
    result = parse_handlers(
        handlers=[Handler(handler_fn, UpdateType.MESSAGE)],
        dependencies=dependencies,
    )
    assert result == {
        "message": OnMessage(
            handlers=[Handler(handler_fn, UpdateType.MESSAGE)],
            dependencies=dependencies,
        )
    }


def test_deps_routers():
    router1 = OnMessage(dependencies={"data1": Provide(provide_data1)})
    router2 = OnMessage(dependencies={"data2": Provide(provide_data2)})
    router3 = OnCallbackQuery()
    result = parse_handlers(
        routers=[router1, router2, router3],
        dependencies={"data1": Provide(provide_data1)}
    )
    assert result == {
        "message": OnMessage(
            routers=[
                OnMessage(dependencies={"data1": Provide(provide_data1)}),
                OnMessage(dependencies={"data2": Provide(provide_data2)}),
            ],
            dependencies={"data1": Provide(provide_data1)},
        ),
        "callback_query": OnCallbackQuery(
            dependencies={"data1": Provide(provide_data1)},
        )
    }



from __future__ import annotations

__all__ = (
    "HandlerGuard",
    "HandlerCallable",
    "HandlerMiddleware",
    "RouterGurad",
    "RouterCallable",
    "RouterMiddleware",
    "Handlers",
    "MiddlewareCallable",
    "UpdateT",
    "UpdateModel",
    "ReplyMarkup",
    "NoneStr",
    "NoneInt",
    "NoneBool",
)

from typing import TYPE_CHECKING
from typing import Any
from typing import Concatenate
from typing import Literal
from typing import ParamSpec
from typing import Protocol
from typing import TypeAlias
from typing import TypedDict
from typing import TypeVar
from typing import runtime_checkable

if TYPE_CHECKING:
    from collections.abc import AsyncGenerator
    from collections.abc import Awaitable
    from collections.abc import Callable

    from .abc import AbstractHandler
    from .bot import Bot
    from .middleware import Middleware
    from .types import BusinessConnection
    from .types import BusinessMessagesDeleted
    from .types import CallbackQuery
    from .types import ChatBoostRemoved
    from .types import ChatBoostUpdated
    from .types import ChatJoinRequest
    from .types import ChatMemberUpdated
    from .types import ChosenInlineResult
    from .types import ForceReply
    from .types import InlineKeyboardMarkup
    from .types import InlineQuery
    from .types import Message
    from .types import MessageReactionCountUpdated
    from .types import MessageReactionUpdated
    from .types import Poll
    from .types import PollAnswer
    from .types import PreCheckoutQuery
    from .types import ReplyKeyboardMarkup
    from .types import ReplyKeyboardRemove
    from .types import ShippingQuery
    from .types import Update
    from .types.abc import TelegramType

    UpdateT = TypeVar(
        "UpdateT",
        bound=(
            CallbackQuery
            | ChatJoinRequest
            | ChatMemberUpdated
            | InlineQuery
            | ChosenInlineResult
            | Message
            | MessageReactionCountUpdated
            | MessageReactionUpdated
            | Poll
            | PollAnswer
            | PreCheckoutQuery
            | ShippingQuery
            | ChatBoostRemoved
            | ChatBoostUpdated
            | BusinessConnection
            | BusinessMessagesDeleted
        )
    )

    ResultModelT = TypeVar(
        "ResultModelT", bound=TelegramType | list | int | str | bool
    )

else:
    UpdateT = TypeVar("UpdateT")
    ResultModelT = TypeVar("ResultModelT")

T = TypeVar("T")
P = ParamSpec("P")
T_contra = TypeVar("T_contra", contravariant=True)

UpdateModel: TypeAlias = (
    "CallbackQuery "
    "| ChatJoinRequest "
    "| ChatMemberUpdated "
    "| InlineQuery "
    "| ChosenInlineResult "
    "| Message "
    "| MessageReactionCountUpdated "
    "| MessageReactionUpdated "
    "| Poll "
    "| PollAnswer "
    "| PreCheckoutQuery "
    "| ShippingQuery"
    "| ChatBoostRemoved"
    "| ChatBoostUpdated"
    "| BusinessConnection"
    "| BusinessMessagesDeleted"
)

ReplyMarkup: TypeAlias = (
    "ForceReply "
    "| InlineKeyboardMarkup "
    "| ReplyKeyboardMarkup "
    "| ReplyKeyboardRemove"
)

NoneStr: TypeAlias = "str | None"
NoneInt: TypeAlias = "int | None"
NoneBool: TypeAlias = "bool | None"

MiddlewareCallable: TypeAlias = "Callable[[T], T]"

RouterCallable: TypeAlias = "Callable[[Update], Awaitable[None]]"
RouterGurad: TypeAlias = "Callable[[Update], Awaitable[None]]"
RouterMiddleware: TypeAlias = "Middleware[RouterCallable] | MiddlewareCallable[RouterCallable]"  # noqa: E501

_HandlerCallable: TypeAlias = "Callable[Concatenate[UpdateT, P], Awaitable[None]]"  # noqa: E501
HandlerCallable: TypeAlias = "_HandlerCallable[UpdateT, Any]"
HandlerCallableType: TypeAlias = "Callable[[UpdateT], Awaitable[None]]"
HandlerGuard: TypeAlias = "Callable[[UpdateT], Awaitable[None]]"
HandlerMiddleware: TypeAlias = (
    "Middleware[HandlerCallableType[UpdateT]] "
    "| MiddlewareCallable[HandlerCallableType[UpdateT]]"
)
HandlerDependency: TypeAlias = (
    "Callable[..., Awaitable[Any]] "
    "| Callable[..., AsyncGenerator[Any, None]] "
    "| type[Any]"
)

Scope: TypeAlias = "Literal['router', 'handler', 'local']"

PollingHook: TypeAlias = "Callable[[Bot], Awaitable[None]]"

FilterGroup: TypeAlias = "Literal['content', 'sender', 'chat']"
FilterPriority: TypeAlias = "dict[FilterGroup, tuple[int, int | tuple[int, int]]]"  # noqa: E501


class Handlers(TypedDict):
    message: AbstractHandler[Message] | None
    edited_message: AbstractHandler[Message] | None
    channel_post: AbstractHandler[Message] | None
    edited_channel_post: AbstractHandler[Message] | None
    message_reaction: AbstractHandler[MessageReactionUpdated] | None
    message_reaction_count: AbstractHandler[MessageReactionCountUpdated] | None
    inline_query: AbstractHandler[InlineQuery] | None
    chosen_inline_result: AbstractHandler[ChosenInlineResult] | None
    callback_query: AbstractHandler[CallbackQuery] | None
    shipping_query: AbstractHandler[ShippingQuery] | None
    pre_checkout_query: AbstractHandler[PreCheckoutQuery] | None
    poll: AbstractHandler[Poll] | None
    poll_answer: AbstractHandler[PollAnswer] | None
    my_chat_member: AbstractHandler[ChatMemberUpdated] | None
    chat_member: AbstractHandler[ChatMemberUpdated] | None
    chat_join_request: AbstractHandler[ChatJoinRequest] | None
    chat_boost: AbstractHandler[ChatBoostUpdated] | None
    removed_chat_boost: AbstractHandler[ChatBoostRemoved] | None


@runtime_checkable
class InputFile(Protocol):
    """File protocol"""

    @property
    def file_name(self) -> str:
        """Name of file."""
        pass

    async def read(self) -> bytes:
        """Returns the file content."""
        pass
